@@ -24,6 +24,12 @@ struct netlink_linearize_ctx {
 	unsigned int		reg_low;
 };
 
+static void netlink_put_register(struct nft_rule_expr *nle,
+				 uint32_t attr, uint32_t reg)
+{
+	nft_rule_expr_set_u32(nle, attr, reg);
+}
+
 static enum nft_registers get_register(struct netlink_linearize_ctx *ctx)
 {
 	if (ctx->reg_low > NFT_REG_MAX)
@@ -57,7 +63,7 @@ static void netlink_gen_payload(struct netlink_linearize_ctx *ctx,
 	struct nft_rule_expr *nle;
 
 	nle = alloc_nft_expr("payload");
-	nft_rule_expr_set_u32(nle, NFT_EXPR_PAYLOAD_DREG, dreg);
+	netlink_put_register(nle, NFT_EXPR_PAYLOAD_DREG, dreg);
 	nft_rule_expr_set_u32(nle, NFT_EXPR_PAYLOAD_BASE,
 			      expr->payload.base - 1);
 	nft_rule_expr_set_u32(nle, NFT_EXPR_PAYLOAD_OFFSET,
@@ -74,7 +80,7 @@ static void netlink_gen_exthdr(struct netlink_linearize_ctx *ctx,
 	struct nft_rule_expr *nle;
 
 	nle = alloc_nft_expr("exthdr");
-	nft_rule_expr_set_u32(nle, NFT_EXPR_EXTHDR_DREG, dreg);
+	netlink_put_register(nle, NFT_EXPR_EXTHDR_DREG, dreg);
 	nft_rule_expr_set_u8(nle, NFT_EXPR_EXTHDR_TYPE,
 			     expr->exthdr.desc->type);
 	nft_rule_expr_set_u32(nle, NFT_EXPR_EXTHDR_OFFSET,
@@ -91,7 +97,7 @@ static void netlink_gen_meta(struct netlink_linearize_ctx *ctx,
 	struct nft_rule_expr *nle;
 
 	nle = alloc_nft_expr("meta");
-	nft_rule_expr_set_u32(nle, NFT_EXPR_META_DREG, dreg);
+	netlink_put_register(nle, NFT_EXPR_META_DREG, dreg);
 	nft_rule_expr_set_u32(nle, NFT_EXPR_META_KEY, expr->meta.key);
 	nft_rule_add_expr(ctx->nlr, nle);
 }
@@ -103,7 +109,7 @@ static void netlink_gen_ct(struct netlink_linearize_ctx *ctx,
 	struct nft_rule_expr *nle;
 
 	nle = alloc_nft_expr("ct");
-	nft_rule_expr_set_u32(nle, NFT_EXPR_CT_DREG, dreg);
+	netlink_put_register(nle, NFT_EXPR_CT_DREG, dreg);
 	nft_rule_expr_set_u32(nle, NFT_EXPR_CT_KEY, expr->ct.key);
 	nft_rule_add_expr(ctx->nlr, nle);
 }
@@ -125,8 +131,8 @@ static void netlink_gen_map(struct netlink_linearize_ctx *ctx,
 	netlink_gen_expr(ctx, expr->map, sreg);
 
 	nle = alloc_nft_expr("lookup");
-	nft_rule_expr_set_u32(nle, NFT_EXPR_LOOKUP_SREG, sreg);
-	nft_rule_expr_set_u32(nle, NFT_EXPR_LOOKUP_DREG, dreg);
+	netlink_put_register(nle, NFT_EXPR_LOOKUP_SREG, sreg);
+	netlink_put_register(nle, NFT_EXPR_LOOKUP_DREG, dreg);
 	nft_rule_expr_set_str(nle, NFT_EXPR_LOOKUP_SET,
 			      expr->mappings->set->handle.set);
 	nft_rule_expr_set_u32(nle, NFT_EXPR_LOOKUP_SET_ID,
@@ -152,7 +158,7 @@ static void netlink_gen_lookup(struct netlink_linearize_ctx *ctx,
 	netlink_gen_expr(ctx, expr->left, sreg);
 
 	nle = alloc_nft_expr("lookup");
-	nft_rule_expr_set_u32(nle, NFT_EXPR_LOOKUP_SREG, sreg);
+	netlink_put_register(nle, NFT_EXPR_LOOKUP_SREG, sreg);
 	nft_rule_expr_set_str(nle, NFT_EXPR_LOOKUP_SET,
 			      expr->right->set->handle.set);
 	nft_rule_expr_set_u32(nle, NFT_EXPR_LOOKUP_SET_ID,
@@ -213,8 +219,8 @@ static void netlink_gen_cmp(struct netlink_linearize_ctx *ctx,
 		zero.len = nld.len;
 
 		nle = alloc_nft_expr("bitwise");
-		nft_rule_expr_set_u32(nle, NFT_EXPR_BITWISE_SREG, sreg);
-		nft_rule_expr_set_u32(nle, NFT_EXPR_BITWISE_DREG, sreg);
+		netlink_put_register(nle, NFT_EXPR_BITWISE_SREG, sreg);
+		netlink_put_register(nle, NFT_EXPR_BITWISE_DREG, sreg);
 		nft_rule_expr_set_u32(nle, NFT_EXPR_BITWISE_LEN, nld.len);
 		nft_rule_expr_set(nle, NFT_EXPR_BITWISE_MASK, &nld.value, nld.len);
 		nft_rule_expr_set(nle, NFT_EXPR_BITWISE_XOR, &zero.value, zero.len);
@@ -230,7 +236,7 @@ static void netlink_gen_cmp(struct netlink_linearize_ctx *ctx,
 	}
 
 	nle = alloc_nft_expr("cmp");
-	nft_rule_expr_set_u32(nle, NFT_EXPR_CMP_SREG, sreg);
+	netlink_put_register(nle, NFT_EXPR_CMP_SREG, sreg);
 	nft_rule_expr_set_u32(nle, NFT_EXPR_CMP_OP,
 			      netlink_gen_cmp_op(expr->op));
 	netlink_gen_data(right, &nld);
@@ -255,7 +261,7 @@ static void netlink_gen_range(struct netlink_linearize_ctx *ctx,
 	netlink_gen_expr(ctx, expr->left, sreg);
 
 	nle = alloc_nft_expr("cmp");
-	nft_rule_expr_set_u32(nle, NFT_EXPR_CMP_SREG, sreg);
+	netlink_put_register(nle, NFT_EXPR_CMP_SREG, sreg);
 	switch (expr->op) {
 	case OP_NEQ:
 		nft_rule_expr_set_u32(nle, NFT_EXPR_CMP_OP,
@@ -275,7 +281,7 @@ static void netlink_gen_range(struct netlink_linearize_ctx *ctx,
 	nft_rule_add_expr(ctx->nlr, nle);
 
 	nle = alloc_nft_expr("cmp");
-	nft_rule_expr_set_u32(nle, NFT_EXPR_CMP_SREG, sreg);
+	netlink_put_register(nle, NFT_EXPR_CMP_SREG, sreg);
 	switch (expr->op) {
 	case OP_NEQ:
 		nft_rule_expr_set_u32(nle, NFT_EXPR_CMP_OP,
@@ -319,15 +325,15 @@ static void netlink_gen_flagcmp(struct netlink_linearize_ctx *ctx,
 	netlink_gen_data(expr->right, &nld2);
 
 	nle = alloc_nft_expr("bitwise");
-	nft_rule_expr_set_u32(nle, NFT_EXPR_BITWISE_SREG, sreg);
-	nft_rule_expr_set_u32(nle, NFT_EXPR_BITWISE_DREG, sreg);
+	netlink_put_register(nle, NFT_EXPR_BITWISE_SREG, sreg);
+	netlink_put_register(nle, NFT_EXPR_BITWISE_DREG, sreg);
 	nft_rule_expr_set_u32(nle, NFT_EXPR_BITWISE_LEN, len);
 	nft_rule_expr_set(nle, NFT_EXPR_BITWISE_MASK, &nld2.value, nld2.len);
 	nft_rule_expr_set(nle, NFT_EXPR_BITWISE_XOR, &nld.value, nld.len);
 	nft_rule_add_expr(ctx->nlr, nle);
 
 	nle = alloc_nft_expr("cmp");
-	nft_rule_expr_set_u32(nle, NFT_EXPR_CMP_SREG, sreg);
+	netlink_put_register(nle, NFT_EXPR_CMP_SREG, sreg);
 	nft_rule_expr_set_u32(nle, NFT_EXPR_CMP_OP, NFT_CMP_NEQ);
 	nft_rule_expr_set(nle, NFT_EXPR_CMP_DATA, nld.value, nld.len);
 	nft_rule_add_expr(ctx->nlr, nle);
@@ -419,8 +425,8 @@ static void netlink_gen_binop(struct netlink_linearize_ctx *ctx,
 	len = div_round_up(expr->len, BITS_PER_BYTE);
 
 	nle = alloc_nft_expr("bitwise");
-	nft_rule_expr_set_u32(nle, NFT_EXPR_BITWISE_SREG, dreg);
-	nft_rule_expr_set_u32(nle, NFT_EXPR_BITWISE_DREG, dreg);
+	netlink_put_register(nle, NFT_EXPR_BITWISE_SREG, dreg);
+	netlink_put_register(nle, NFT_EXPR_BITWISE_DREG, dreg);
 	nft_rule_expr_set_u32(nle, NFT_EXPR_BITWISE_LEN, len);
 
 	netlink_gen_raw_data(mask, expr->byteorder, len, &nld);
@@ -442,7 +448,7 @@ static enum nft_byteorder_ops netlink_gen_unary_op(enum ops op)
 	case OP_HTON:
 		return NFT_BYTEORDER_HTON;
 	case OP_NTOH:
-		return NFT_BYTEORDER_HTON;
+		return NFT_BYTEORDER_NTOH;
 	default:
 		BUG("invalid unary operation %u\n", op);
 	}
@@ -457,8 +463,8 @@ static void netlink_gen_unary(struct netlink_linearize_ctx *ctx,
 	netlink_gen_expr(ctx, expr->arg, dreg);
 
 	nle = alloc_nft_expr("byteorder");
-	nft_rule_expr_set_u32(nle, NFT_EXPR_BYTEORDER_SREG, dreg);
-	nft_rule_expr_set_u32(nle, NFT_EXPR_BYTEORDER_DREG, dreg);
+	netlink_put_register(nle, NFT_EXPR_BYTEORDER_SREG, dreg);
+	netlink_put_register(nle, NFT_EXPR_BYTEORDER_DREG, dreg);
 	nft_rule_expr_set_u32(nle, NFT_EXPR_BYTEORDER_LEN,
 			      expr->len / BITS_PER_BYTE);
 	nft_rule_expr_set_u32(nle, NFT_EXPR_BYTEORDER_SIZE,
@@ -476,7 +482,7 @@ static void netlink_gen_immediate(struct netlink_linearize_ctx *ctx,
 	struct nft_data_linearize nld;
 
 	nle = alloc_nft_expr("immediate");
-	nft_rule_expr_set_u32(nle, NFT_EXPR_IMM_DREG, dreg);
+	netlink_put_register(nle, NFT_EXPR_IMM_DREG, dreg);
 	netlink_gen_data(expr, &nld);
 	switch (expr->ops->type) {
 	case EXPR_VALUE:
@@ -561,7 +567,7 @@ static void netlink_gen_meta_stmt(struct netlink_linearize_ctx *ctx,
 	release_register(ctx);
 
 	nle = alloc_nft_expr("meta");
-	nft_rule_expr_set_u32(nle, NFT_EXPR_META_SREG, sreg);
+	netlink_put_register(nle, NFT_EXPR_META_SREG, sreg);
 	nft_rule_expr_set_u32(nle, NFT_EXPR_META_KEY, stmt->meta.key);
 	nft_rule_add_expr(ctx->nlr, nle);
 }
@@ -576,17 +582,20 @@ static void netlink_gen_log_stmt(struct netlink_linearize_ctx *ctx,
 		nft_rule_expr_set_str(nle, NFT_EXPR_LOG_PREFIX,
 				      stmt->log.prefix);
 	}
-	if (stmt->log.group) {
+	if (stmt->log.flags & STMT_LOG_GROUP) {
 		nft_rule_expr_set_u16(nle, NFT_EXPR_LOG_GROUP,
 				      stmt->log.group);
-	}
-	if (stmt->log.snaplen) {
-		nft_rule_expr_set_u32(nle, NFT_EXPR_LOG_SNAPLEN,
-				      stmt->log.snaplen);
-	}
-	if (stmt->log.qthreshold) {
-		nft_rule_expr_set_u16(nle, NFT_EXPR_LOG_QTHRESHOLD,
-				      stmt->log.qthreshold);
+		if (stmt->log.flags & STMT_LOG_SNAPLEN)
+			nft_rule_expr_set_u32(nle, NFT_EXPR_LOG_SNAPLEN,
+					      stmt->log.snaplen);
+		if (stmt->log.flags & STMT_LOG_QTHRESHOLD)
+			nft_rule_expr_set_u16(nle, NFT_EXPR_LOG_QTHRESHOLD,
+					      stmt->log.qthreshold);
+	} else {
+		if (stmt->log.flags & STMT_LOG_LEVEL) {
+			nft_rule_expr_set_u32(nle, NFT_EXPR_LOG_LEVEL,
+					      stmt->log.level);
+		}
 	}
 	nft_rule_add_expr(ctx->nlr, nle);
 }
@@ -609,7 +618,10 @@ static void netlink_gen_reject_stmt(struct netlink_linearize_ctx *ctx,
 
 	nle = alloc_nft_expr("reject");
 	nft_rule_expr_set_u32(nle, NFT_EXPR_REJECT_TYPE, stmt->reject.type);
-	nft_rule_expr_set_u8(nle, NFT_EXPR_REJECT_CODE, 0);
+	if (stmt->reject.icmp_code != -1)
+		nft_rule_expr_set_u8(nle, NFT_EXPR_REJECT_CODE,
+				     stmt->reject.icmp_code);
+
 	nft_rule_add_expr(ctx->nlr, nle);
 }
 
@@ -628,6 +640,9 @@ static void netlink_gen_nat_stmt(struct netlink_linearize_ctx *ctx,
 	family = nft_rule_attr_get_u32(ctx->nlr, NFT_RULE_ATTR_FAMILY);
 	nft_rule_expr_set_u32(nle, NFT_EXPR_NAT_FAMILY, family);
 
+	if (stmt->nat.flags != 0)
+		nft_rule_expr_set_u32(nle, NFT_EXPR_NAT_FLAGS, stmt->nat.flags);
+
 	if (stmt->nat.addr) {
 		amin_reg = get_register(ctx);
 		registers++;
@@ -638,14 +653,14 @@ static void netlink_gen_nat_stmt(struct netlink_linearize_ctx *ctx,
 
 			netlink_gen_expr(ctx, stmt->nat.addr->left, amin_reg);
 			netlink_gen_expr(ctx, stmt->nat.addr->right, amax_reg);
-			nft_rule_expr_set_u32(nle, NFT_EXPR_NAT_REG_ADDR_MIN,
-					      amin_reg);
-			nft_rule_expr_set_u32(nle, NFT_EXPR_NAT_REG_ADDR_MAX,
-					      amax_reg);
+			netlink_put_register(nle, NFT_EXPR_NAT_REG_ADDR_MIN,
+					     amin_reg);
+			netlink_put_register(nle, NFT_EXPR_NAT_REG_ADDR_MAX,
+					     amax_reg);
 		} else {
 			netlink_gen_expr(ctx, stmt->nat.addr, amin_reg);
-			nft_rule_expr_set_u32(nle, NFT_EXPR_NAT_REG_ADDR_MIN,
-					      amin_reg);
+			netlink_put_register(nle, NFT_EXPR_NAT_REG_ADDR_MIN,
+					     amin_reg);
 		}
 
 	}
@@ -660,13 +675,73 @@ static void netlink_gen_nat_stmt(struct netlink_linearize_ctx *ctx,
 
 			netlink_gen_expr(ctx, stmt->nat.proto->left, pmin_reg);
 			netlink_gen_expr(ctx, stmt->nat.proto->right, pmax_reg);
-			nft_rule_expr_set_u32(nle, NFT_EXPR_NAT_REG_PROTO_MIN,
-					      pmin_reg);
-			nft_rule_expr_set_u32(nle, NFT_EXPR_NAT_REG_PROTO_MAX,
-					      pmax_reg);
+			netlink_put_register(nle, NFT_EXPR_NAT_REG_PROTO_MIN,
+					     pmin_reg);
+			netlink_put_register(nle, NFT_EXPR_NAT_REG_PROTO_MAX,
+					     pmax_reg);
 		} else {
 			netlink_gen_expr(ctx, stmt->nat.proto, pmin_reg);
-			nft_rule_expr_set_u32(nle, NFT_EXPR_NAT_REG_PROTO_MIN,
+			netlink_put_register(nle, NFT_EXPR_NAT_REG_PROTO_MIN,
+					     pmin_reg);
+		}
+	}
+
+	while (registers > 0) {
+		release_register(ctx);
+		registers--;
+	}
+
+	nft_rule_add_expr(ctx->nlr, nle);
+}
+
+static void netlink_gen_masq_stmt(struct netlink_linearize_ctx *ctx,
+				  const struct stmt *stmt)
+{
+	struct nft_rule_expr *nle;
+
+	nle = alloc_nft_expr("masq");
+	if (stmt->masq.flags != 0)
+		nft_rule_expr_set_u32(nle, NFT_EXPR_MASQ_FLAGS,
+				      stmt->masq.flags);
+
+	nft_rule_add_expr(ctx->nlr, nle);
+}
+
+static void netlink_gen_redir_stmt(struct netlink_linearize_ctx *ctx,
+				   const struct stmt *stmt)
+{
+	struct nft_rule_expr *nle;
+	enum nft_registers pmin_reg, pmax_reg;
+	int registers = 0;
+
+	nle = alloc_nft_expr("redir");
+
+	if (stmt->redir.flags != 0)
+		nft_rule_expr_set_u32(nle, NFT_EXPR_REDIR_FLAGS,
+				      stmt->redir.flags);
+
+	if (stmt->redir.proto) {
+		pmin_reg = get_register(ctx);
+		registers++;
+
+		if (stmt->redir.proto->ops->type == EXPR_RANGE) {
+			pmax_reg = get_register(ctx);
+			registers++;
+
+			netlink_gen_expr(ctx, stmt->redir.proto->left,
+					 pmin_reg);
+			netlink_gen_expr(ctx, stmt->redir.proto->right,
+					 pmax_reg);
+			nft_rule_expr_set_u32(nle,
+					      NFT_EXPR_REDIR_REG_PROTO_MIN,
+					      pmin_reg);
+			nft_rule_expr_set_u32(nle,
+					      NFT_EXPR_REDIR_REG_PROTO_MAX,
+					      pmax_reg);
+		} else {
+			netlink_gen_expr(ctx, stmt->redir.proto, pmin_reg);
+			nft_rule_expr_set_u32(nle,
+					      NFT_EXPR_REDIR_REG_PROTO_MIN,
 					      pmin_reg);
 		}
 	}
@@ -684,21 +759,27 @@ static void netlink_gen_queue_stmt(struct netlink_linearize_ctx *ctx,
 {
 	struct nft_rule_expr *nle;
 	uint16_t total_queues;
+	mpz_t low, high;
+
+	mpz_init2(low, 16);
+	mpz_init2(high, 16);
+	if (stmt->queue.queue != NULL) {
+		range_expr_value_low(low, stmt->queue.queue);
+		range_expr_value_high(high, stmt->queue.queue);
+	}
+	total_queues = mpz_get_uint16(high) - mpz_get_uint16(low) + 1;
 
 	nle = alloc_nft_expr("queue");
-
-	nft_rule_expr_set_u16(nle, NFT_EXPR_QUEUE_NUM,
-			      stmt->queue.from);
-
-	total_queues = stmt->queue.to - stmt->queue.from;
-	nft_rule_expr_set_u16(nle, NFT_EXPR_QUEUE_TOTAL,
-			      total_queues + 1);
-
+	nft_rule_expr_set_u16(nle, NFT_EXPR_QUEUE_NUM, mpz_get_uint16(low));
+	nft_rule_expr_set_u16(nle, NFT_EXPR_QUEUE_TOTAL, total_queues);
 	if (stmt->queue.flags) {
 		nft_rule_expr_set_u16(nle, NFT_EXPR_QUEUE_FLAGS,
 				      stmt->queue.flags);
 	}
 	nft_rule_add_expr(ctx->nlr, nle);
+
+	mpz_clear(low);
+	mpz_clear(high);
 }
 
 static void netlink_gen_ct_stmt(struct netlink_linearize_ctx *ctx,
@@ -712,7 +793,7 @@ static void netlink_gen_ct_stmt(struct netlink_linearize_ctx *ctx,
 	release_register(ctx);
 
 	nle = alloc_nft_expr("ct");
-	nft_rule_expr_set_u32(nle, NFT_EXPR_CT_SREG, sreg);
+	netlink_put_register(nle, NFT_EXPR_CT_SREG, sreg);
 	nft_rule_expr_set_u32(nle, NFT_EXPR_CT_KEY, stmt->ct.key);
 	nft_rule_add_expr(ctx->nlr, nle);
 }
@@ -737,6 +818,10 @@ static void netlink_gen_stmt(struct netlink_linearize_ctx *ctx,
 		return netlink_gen_reject_stmt(ctx, stmt);
 	case STMT_NAT:
 		return netlink_gen_nat_stmt(ctx, stmt);
+	case STMT_MASQ:
+		return netlink_gen_masq_stmt(ctx, stmt);
+	case STMT_REDIR:
+		return netlink_gen_redir_stmt(ctx, stmt);
 	case STMT_QUEUE:
 		return netlink_gen_queue_stmt(ctx, stmt);
 	case STMT_CT:
@@ -746,8 +831,8 @@ static void netlink_gen_stmt(struct netlink_linearize_ctx *ctx,
 	}
 }
 
-int netlink_linearize_rule(struct netlink_ctx *ctx, struct nft_rule *nlr,
-			   const struct rule *rule)
+void netlink_linearize_rule(struct netlink_ctx *ctx, struct nft_rule *nlr,
+			    const struct rule *rule)
 {
 	struct netlink_linearize_ctx lctx;
 	const struct stmt *stmt;
@@ -760,5 +845,4 @@ int netlink_linearize_rule(struct netlink_ctx *ctx, struct nft_rule *nlr,
 		netlink_gen_stmt(&lctx, stmt);
 
 	netlink_dump_rule(nlr);
-	return 0;
 }
